# SPDX-License-Identifier: MIT

status_check_last_update() {
	# ugly: optimal / ts are defined in ctrl_status and used there
	optimal="$(cat /var/log/swupdate/last_update 2>/dev/null)"
	ts="${optimal##* }"
	optimal="${optimal% *}"

	if [ -z "$optimal" ]; then
		info "rollback-status: fresh install"
		return
	fi

	# rolled back, return early
	[ "$optimal" = "$rootpart" ] || return

	if fw_printenv_nowarn upgrade_available | grep -qx 'upgrade_available=1'; then
		info "rollback-status: optimal"
	else
		info 'rollback-status: no fallback'
	fi
}

status_check_public_swu_pem() {
	local crt onetime

	# If onetime-public certificate is present it is the first in
	# swupdate.pem, so we do not need to iterate all certs
	crt="$(openssl x509 -noout -in /etc/swupdate.pem -pubkey  2>/dev/null \
		| sed -e '/-----/d' | tr -d '\n')"
	onetime="MFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAEYTN7NghmISesYQ1dnby5YkocLAe2/EJ8OTXkx/xGhBVlJ57eGOovtPORd/JMkA6lWI0N/pD5p6eUGcwrQvRtsw=="

	[ "$crt" != "$onetime" ]
}

status_check_modules_installed() {
	# just check if modules dir exists
	test -d "/lib/modules/$(uname -r)" && return

	# if not, verify we have any module compiled in if we can.
	# absence of /proc/config.gz or presence of 'as module' in it
	# are considered failures.
	[ -e /proc/config.gz ] && ! zcat /proc/config.gz | grep -q "=m"
}

status_help() {
	echo "Usage: abos-ctrl [common options] status [extra options]"
	echo
	echo "  --onetime-cert    only check for onetime certificate presence"
	echo "  --rollback        only check for rollback status"
	echo "both options imply quiet"
	echo
	echo "see abos-ctrl --help for common options"
}

ctrl_status() {
	local fail=""
	# used with status_check_last_update
	local optimal ts

	dev_from_current

	while [ "$#" -gt 0 ]; do
		case "$1" in
		"--onetime-cert")
			status_check_public_swu_pem
			return
			;;
		"--rollback")
			debug=$((debug-1))
			status_check_last_update
			return
			;;
		"-q"|"--quiet")
			debug=$((debug-1))
			;;
		"-v"|"--verbose")
			debug=$((debug+1))
			;;
		*)
			warning "Invalid option $1"
			status_help >&2
			exit 1
			;;
		esac
	done


	info "Currently booted on $rootpart"

	# compare with /var/log/swupdate/last_update
	if ! status_check_last_update; then
		warning "Currently running on non-latest version (expected $optimal${ts:+ installed on $(date -d "@$ts")})"
		info "rollback-status: rolled back"
		fail=1
	fi

	if ! status_check_public_swu_pem; then
		warning "swupdate onetime public certificate is present, anyone can access this device" \
			"Please install initial_setup.swu (from mkswu --init)," \
			"or remove the first certificate from /etc/swupdate.pem"
	fi

	if ! status_check_modules_installed; then
		warning "This kernel requires modules, but modules were not found in /lib/modules/$(uname -r)." \
			"Please check your kernel installation or re-install default kernel with" \
			"'abos-ctrl restore-kernel'"
	fi

	[ -z "$fail" ] || exit 1
}
